//
//  DataManager.swift
//  CollectMe
//
//  Created by Oscar Fernandez Ardanuy on 14/04/16.
//  Copyright © 2016 dorhus. All rights reserved.
//

import UIKit
import CoreData

class DataManager {

    // COLLECTIONS -------------------------------------------------------------------
    
    private static var myStoredCollections:[MyCollection] = []
    private static var myStoredItems:[MyItem] = []
    private static var myCollectionsWasModified:Bool = false
    private static var myItemsWasModified:Bool = false
    
    // completion nil when fail
    static func collectionExists(collection: MyCollection, completion: (Bool?) -> (Void)) {
        self.getAllMyCollection { (collectionList: [MyCollection]?) -> (Void) in
            if collectionList != nil {
                completion(collectionList!.contains(collection))
            } else {
                completion(nil)
            }
        }
    }
    
    // completion nil when fail
    static func getAllMyCollection(completion: ([MyCollection]?) -> (Void)) {
        
        if (myCollectionsWasModified || myStoredCollections.isEmpty) {
            myCollectionsWasModified = false
            
            let managedContext = getManagedObjectContext()
            let fetchRequest = NSFetchRequest(entityName: MY_COLLECTION_ENTITY_NAME)
            
            do {
                let results = try managedContext.executeFetchRequest(fetchRequest)
                let allMyCollections = results as! [MyCollection]
                
                myStoredCollections = allMyCollections
                completion(allMyCollections)
                
            } catch let error as NSError {
                print("getAllMyCollection Could not load \(error), \(error.userInfo)")
                completion(nil)
            }
        
        } else {
            completion(myStoredCollections)
        }
        
        //print("DataManager.getAllMyCollection")
        //print(myStoredCollections)
        //print(" ")
    }
    
    // completion nil when fail
    static func getMyCollection(from item: MyItem, completion: (MyCollection?) -> (Void)) {
        getAllMyCollection { (allCollections:[MyCollection]?) -> (Void) in
            if allCollections != nil {
                for collection in allCollections! {
                    if let itemList = collection.items {
                        if itemList.containsObject(item) {
                            completion(collection)
                            break
                        }
                    }
                }
            } else {
                completion(nil)
            }
        }
    }
    
    static func getMyCollectionsCount(completion: (Int) -> (Void)) {
        getAllMyCollection { (allCollections:[MyCollection]?) -> (Void) in
            if (allCollections != nil) {
                completion(allCollections!.count)
            } else {
                completion(0)
            }
        }
    }
    
    // completion false when fail
    static func saveNewMyCollection(name: String, info: String?, full: Bool, color: String, completion: (Bool) -> (Void)) {
        let managedContext = getManagedObjectContext()
        
        let entity = NSEntityDescription.entityForName(MY_COLLECTION_ENTITY_NAME, inManagedObjectContext: managedContext)
        let myCollection = MyCollection(entity: entity!, insertIntoManagedObjectContext: managedContext)
        
        myCollection.name = name
        myCollection.info = info
        myCollection.setIsFull(full)
        myCollection.color = color
        
        do {
            try managedContext.save()
            //myCollectionsWasModified = true
            myStoredCollections.append(myCollection)
            
            completion(true)
        } catch let error as NSError {
            print("DataManager.saveNewMyCollection Could not save \(error), \(error.userInfo)")
            completion(false)
        }
    }
    
    // completion nil when fail
    static func saveAndReturnNewMyCollection(name: String, info: String?, full: Bool, completion: (MyCollection?) -> (Void)) {
        let managedContext = getManagedObjectContext()
        
        let entity = NSEntityDescription.entityForName(MY_COLLECTION_ENTITY_NAME, inManagedObjectContext: managedContext)
        let myCollection = MyCollection(entity: entity!, insertIntoManagedObjectContext: managedContext)
        
        myCollection.name = name
        myCollection.info = info
        myCollection.setIsFull(full)
        myCollection.color = CollectionsColor.colors[0]
        
        do {
            try managedContext.save()
            //myCollectionsWasModified = true
            myStoredCollections.append(myCollection)
            
            completion(myCollection)
        } catch let error as NSError {
            print("DataManager.saveNewMyCollection Could not save \(error), \(error.userInfo)")
            completion(nil)
        }
    }
    
    // completion false when fail
    static func updateMyCollection(myCollection: MyCollection, completion: (Bool) -> (Void)) {
        
        do {
            try myCollection.managedObjectContext?.save()
            //myCollectionsWasModified = true
            //myStoredCollections[myStoredCollections.indexOf(myCollection)!] = myCollection
            
            completion(true)
        } catch let error as NSError {
            print("DataManager.updateMyCollection Could not update \(error), \(error.userInfo)")
            completion(false)
        }
    }
    
    // completion false when fail
    static func deleteMyCollection(myCollection: MyCollection, completion: (Bool) -> (Void)) {
        self.myItemsWasModified = true
        
        let managedContext = getManagedObjectContext()
        let indexCollection = myStoredCollections.indexOf(myCollection)
        managedContext.deleteObject(myCollection)
        
        do {
            try managedContext.save()
            //myCollectionsWasModified = true
            myStoredCollections.removeAtIndex(indexCollection!)
            
            completion(true)
        } catch let error as NSError {
            print("DataManager.deleteMyCollection Could not delete \(error), \(error.userInfo)")
            completion(false)
        }
    }

    
    // ITEMS -------------------------------------------------------------------
    
    // completion nil when fail
    static func itemExists(item: MyItem, completion: (Bool?) -> (Void)) {
        self.getAllItems { (itemsList: [MyItem]?) -> (Void) in
            if itemsList != nil {
                completion(itemsList!.contains(item))
            } else {
                completion(nil)
            }
        }
    }
    
    // completion nil when fail
    static func getWishList(completion: ([MyItem]?) -> (Void)) {
        self.getAllMyCollection { (allCollections:[MyCollection]?) -> (Void) in
            if (allCollections != nil){
                var wishList:[MyItem] = []
                
                for oneCollection:MyCollection in allCollections! {
                    if (oneCollection.items?.count > 0) {
                        for (index: _, element: item) in oneCollection.items!.enumerate() {
                            let oneItem = item as! MyItem
                            if (oneItem.isWanted()) {
                                wishList.append(oneItem)
                            }
                        }
                    }
                }
                completion(wishList)
            } else {
                completion(nil)
            }
        }
    }
    
    static func getMyWishedItemsCount(completion: (Int) -> (Void)) {
        getWishList { (allWishedItems:[MyItem]?) -> (Void) in
            if (allWishedItems != nil) {
                completion(allWishedItems!.count)
            } else {
                completion(0)
            }
        }
    }
    
    // completion nil when fail
    static func getAllItems(completion: ([MyItem]?) -> (Void)) {
        
        if (myItemsWasModified || self.myStoredItems.isEmpty) {
            self.getAllMyCollection { (allCollections:[MyCollection]?) -> (Void) in
                if (allCollections != nil){
                    var itemsArray:[MyItem] = []
                    
                    for oneCollection:MyCollection in allCollections! {
                        itemsArray += oneCollection.items?.allObjects as! [MyItem]
                    }
                    self.myStoredItems = itemsArray
                    self.myItemsWasModified = false
                    
                    completion(itemsArray)
                } else {
                    completion(nil)
                }
            }
        } else {
            completion(self.myStoredItems)
        }
    }
    
    static func getMyItemsCount(completion: (Int) -> (Void)) {
        getAllItems { (allItems:[MyItem]?) -> (Void) in
            if (allItems != nil) {
                completion(allItems!.count)
            } else {
                completion(0)
            }
        }
    }
    
    // completion false when fail
    static func saveNewMyItem(inside myCollectionNode: MyCollection, name: String,
        desc: String?, notes: String?, barcode: String?,
        shop: String?, type: String?, whereIs: String?,
        author: String?, currencyPaid: String?, amountPaid: String?,
        condition: Int?, personalValue: Int?, serialOrder: String?,
        serialSize: String?, isComplete: Bool?, isWanted: Bool?,
        withBox: Bool?, releaseDate: NSDate?, images: CarouselPictures,
        completion: (Bool) -> (Void)) {
            
        self.myItemsWasModified = true
        
        let managedContext = getManagedObjectContext()
        
        let entity = NSEntityDescription.entityForName(MY_ITEM_ENTITY_NAME, inManagedObjectContext: managedContext)
        let myItem = MyItem(entity: entity!, insertIntoManagedObjectContext: managedContext)
        
        myItem.collection_node = myCollectionNode
        myItem.name_title = name
        myItem.desc = desc
        myItem.notes = notes
        myItem.barcode = barcode
        myItem.shop = shop
        myItem.type_class = type
        myItem.where_is = whereIs
        myItem.author = author
        myItem.currency_paid = currencyPaid
        myItem.amount_paid = amountPaid
        myItem.status_condition = condition
        myItem.personal_value = personalValue
        myItem.serial_order = serialOrder
        myItem.serial_size = serialSize
        if (isComplete != nil) {
            myItem.setIsComplete(isComplete!)
        }
        if (isWanted != nil) {
            myItem.setIsWanted(isWanted!)
        }
        if (withBox != nil) {
            myItem.setHasBox(withBox!)
        }
        myItem.release_date = releaseDate
    
        //images
        let newPics = preparedNewPictures(images.getNewPicturesArray())
        addPictures(inside: myItem, images: newPics)
        
        do {
            //print("DataManager.saveNewMyItem BEFORE")
            //print(myStoredCollections)
            //print(" ")
            
            try managedContext.save()
            
            // Class Cache
            let myCollectionItemsSet = myStoredCollections[myStoredCollections.indexOf(myCollectionNode)!].items as! NSMutableSet
            myCollectionItemsSet.addObject(myItem)
            
            
            //print("DataManager.saveNewMyItem AFTER")
            //print(myStoredCollections)
            //print(" ")
            
            completion(true)
        } catch let error as NSError {
            print("DataManager.saveNewMyItem Could not save \(error), \(error.userInfo)")
            completion(false)
        }
    }
    
    // completion false when fail
    static func updateMyItem(this item: MyItem, images: CarouselPictures, completion: (Bool) -> (Void)) {
        self.myItemsWasModified = true
        
        do {
            //images
            let newPics = preparedNewPictures(images.getNewPicturesArray())
            addPictures(inside: item, images: newPics)
            
            let deletedOldPics = images.getDeletedOldPicturesArray()
            deletePictures(deletedOldPics)
            
            try item.managedObjectContext?.save()
            
            completion(true)
        } catch let error as NSError {
            print("DataManager.updateMyItem Could not update \(error), \(error.userInfo)")
            completion(false)
        }
    }
    static func updateMyItem_old(this item: MyItem, completion: (Bool) -> (Void)) {
        
        do {
            try item.managedObjectContext?.save()
            
            completion(true)
        } catch let error as NSError {
            print("DataManager.updateMyItem Could not update \(error), \(error.userInfo)")
            completion(false)
        }
    }
    
    // completion false when fail
    static func deleteMyItem(myItem: MyItem, completion: (Bool) -> (Void)) {
        self.myItemsWasModified = true
        
        let managedContext = getManagedObjectContext()
        managedContext.deleteObject(myItem)
        
        do {
            try managedContext.save()
            
            completion(true)
        } catch let error as NSError {
            print("DataManager.deleteMyItem Could not delete \(error), \(error.userInfo)")
            completion(false)
        }
    }
    
    
    // PICTURES -------------------------------------------------------------------
    private static func preparedNewPictures(newPicturesArray: [PictureWithThumbnail]) -> [PicturesThumbnail] {
        let managedContext = getManagedObjectContext()
        var newThumbnails: [PicturesThumbnail] = []
        
        for newPic in newPicturesArray {
            // Thumbnail Picture
            let entityThumbnails = NSEntityDescription.entityForName(PICTURES_THUMBNAIL_ENTITY_NAME, inManagedObjectContext: managedContext)
            let pThumbnail = PicturesThumbnail(entity: entityThumbnails!, insertIntoManagedObjectContext: managedContext)
            pThumbnail.image_data = newPic.thumbnail
            pThumbnail.id = newPic.timestamp
            
            //print(timestamp)
            
            // Full Resolution Picture
            let entityFullResPictures = NSEntityDescription.entityForName(PICTURES_FULL_RES_ENTITY_NAME, inManagedObjectContext: managedContext)
            let pFullRes = PicturesFullRes(entity: entityFullResPictures!, insertIntoManagedObjectContext: managedContext)
            pFullRes.image_data = newPic.picture
            pFullRes.thumbnail = pThumbnail
            
            newThumbnails.append(pThumbnail)
        }
        
        return newThumbnails
    }
    
    private static func addPictures(inside item: MyItem, images: [PicturesThumbnail]) {
        for picture in  images {
            picture.item = item
        }
    }
    
    private static func deletePictures(oldPicsArray: [PicturesThumbnail]) {
        let managedContext = getManagedObjectContext()
        
        for oldPic in oldPicsArray {
            managedContext.deleteObject(oldPic)
            
            do {
                try managedContext.save()
                
            } catch let error as NSError {
                print("DataManager.deletePictures Could not delete \(error), \(error.userInfo)")
            }
        }
    }
    
    
    
    
    static func testReturn(completion: (Bool) -> (Void)) {
        completion(true)
    }
    
    private static func getManagedObjectContext() -> NSManagedObjectContext{
        let appDelegate = UIApplication.sharedApplication().delegate as! AppDelegate
        let managedContext = appDelegate.managedObjectContext
        
        return managedContext
    }
    
    
}
